% This file is part of YWEB package.
% ----------------------------------------------------------------
% $Source: /home/web/cweb-iso8859/yweb/examples/RCS/hugs-par.w,v $
% $Date: 1998/04/17 17:19:49 $
% $Revision: 1.3 $
% ----------------------------------------------------------------

\def\title{hugs parser (version 1.4)}
\def\topofcontents{\null\vfill
  \leftline{\sc\today, hour \hours}
  \vskip1.5in
  \centerline{\titlefont The Hugs parser (selection from the version 1.4)}
  \vfill}
\def\botofcontents{\vfill
\noindent An arbitrary selection by W{\l}odek Bzyl (March 1998)\par}
\pageno=\contentspagenumber \advance\pageno by 1
\let\maybe=\iftrue

% Forward declarations:
@s start YACC
@s topDecl YACC
@s fixDecl YACC
@s begin YACC
@s type YACC
@s ttype YACC
@s EXPR YACC
@s SCRIPT YACC
@s qcon YACC
@s qconid YACC

@s Int int
@s Cell int
@s String int
@s Void int
@s Syntax int
@s Pair int
@s List int
@s local int

@* Introduction.
This file serves as the test for \.{YWEAVE} processor and not as an
example of literate programming.  It contains fragments of the
\.{Hugs} parser contained in the file \.{parser.y} from the \.{Hugs}
distribution. The \CEE/ code contained there is omitted.  Also note
that \.{CWEAVE} (and therefore \.{YWEAVE}) does not handle ANSII
prototyping declarations.

@f line YACC /* make \\{line} unreserved word */
@f start YACC /* make \\{start} unreserved word */

@* The overall layout of a YACC source.

@c
%{
@<\CEE/ declarations@>
%}
@<\YACC/ declarations@>
%%
@<Grammar rules and actions@>
%%
@<Additional \CEE/ code @>

@ For the purposes of reasonably portable garbage collection, it is
necessary to simulate the \YACC/ stack on the \.{Hugs} stack to keep track
of all intermediate constructs.  The lexical analyser pushes a token
onto the stack for each token that is found, with these elements being
removed as reduce actions are performed, taking account of look-ahead
tokens as described by |gcShadow()| below.

Of the non-terminals used below, only |start|, |topDecl|, |fixDecl| and |begin|
do not leave any values on the \.{Hugs} stack.  The same is true for the
terminals |EXPR| and |SCRIPT|.  At the end of a successful parse, there
should only be one element left on the stack, containing the result of
the parse.

@<\CEE/ declarations@>=
@h
#ifndef lint
#define lint
#endif

@* The Hugs grammar.

@<\YACC/ declarations@>=
%token EXPR       SCRIPT
%token CASEXP     OF         DATA       TYPE       IF
%token THEN       ELSE       WHERE      LET        IN
%token INFIX      INFIXL     INFIXR     PRIMITIVE  TNEWTYPE
%token DEFAULT    DERIVING   DO         TCLASS     TINSTANCE
%token TRUNST     REPEAT
%token VAROP      VARID      NUMLIT     CHARLIT    STRINGLIT
%token CONOP      CONID
%token QVAROP     QVARID     QCONOP     QCONID
%token RECSELID
%token COCO       '='        UPTO       '@@'       '\\'
%token '|'        '-'        FROM       ARROW      '~'
%token '!'        IMPLIES    '('        ','        ')'
%token '['        ';'        ']'        '`'
%token MODULETOK  IMPORT     HIDING     QUALIFIED  ASMOD
%token NEEDPRIMS

@ Haskell module header/import parsing.  Syntax for Haskell modules
(module headers and imports) is parsed but most of it is ignored.
However, module names in import declarations are used, of course, if
import chasing is turned on.

@<Grammar rules and actions@>=
@<Top level script/module structure@>@;
@<Exports@>@;
@<Import declarations@>@;

@ To implement the Haskell module system, we have to keep track of the
current module.  We rely on the use of LALR parsing to ensure that
this side effect happens before any declarations within the module.

@<Top level script/module structure@>=
start     : EXPR exp wherePart          {|inputExpr = letrec($3,$2); sp-=2;|}
	  | SCRIPT topModule            {|valDefns  = $2;            sp-=1;|}
	  | error                       {|syntaxError("input");|}
	  ;
topModule : startMain begin modBody end {
	setExportList(singleton(ap(MODULEENT,mkCon(module(currentModule).text))));
                                         $$ = gc3($3);
                                        }
	  | MODULETOK modname expspec WHERE '{' modBody end
					{ setExportList($3);   $$ = gc7($6); }
	  | MODULETOK error             {|syntaxError("module definition");|}
	  ;
startMain :                             {|startModule(conMain); $$ = gc0(NIL);|}
          ;
modname   : CONID                       {|startModule($1); $$ = gc1(NIL);|}
          ;
modid     : CONID			{|$$ = $1;|}
	  | STRINGLIT			{ static String scriptFile;
 		String modName = findPathname(scriptFile,textToStr(textOf($1)));
                                          if (modName) { /* fillin pathname if known */
					      $$ = mkStr(findText(modName));
					  } else {
					      $$ = $1;
					  }
                                        }
	  ;
modBody   : topDecls			{|$$ = $1;|}
	  | fixDecls ';' topDecls	{|$$ = gc3($3);|}
	  | impDecls chase		{|$$ = gc2(NIL);|}
	  | impDecls ';' chase topDecls	{|$$ = gc4($4);|}
	  | impDecls ';' chase fixDecls ';' topDecls
					{|$$ = gc6($6);|}
	  ;

@ The |qcon| should be |qconid|.  Relaxing the rule lets us explicitly
export from the \.{Prelude}.

@<Exports@>=
expspec   :                             {|$$ = gc0(exportSelf());|}
	  | '(' ')'                     {|$$ = gc2(NIL);|}
	  | '(' exports ')'             {|$$ = gc3($2);|}
	  | '(' exports ',' ')'         {|$$ = gc4($2);|}
	  ;
exports   : exports ',' export          {|$$ = gc3(cons($3,$1));|}
	  | export                      {|$$ = gc1(singleton($1));|}
	  ;
export	  : qvar			{|$$ = $1;|}
	  | qcon			{|$$ = $1;|}
	  | qconid '(' UPTO ')'		{|$$ = gc4(pair($1,DOTDOT));|}
	  | qconid '(' qnames ')'	{|$$ = gc4(pair($1,$3));|}
	  | MODULETOK modid		{|$$ = gc2(ap(MODULEENT,$2));|}
	  ;
qnames    :                             {|$$ = gc0(NIL);|}
	  | ','                         {|$$ = gc1(NIL);|}
	  | qnames1                     {|$$ = $1;|}
	  | qnames1 ','                 {|$$ = gc2($1);|}
	  ;
qnames1   : qnames1 ',' qname           {|$$ = gc3(cons($3,$1));|}
	  | qname                       {|$$ = gc1(singleton($1));|}
	  ;
qname     : qvar                        {|$$ = $1;|}
	  | qcon                        {|$$ = $1;|}
	  ;

@ Note that qualified import ignores the import list.

@<Import declarations@>=
impDecls  : impDecls ';' impDecl	{|imps = cons($3,imps); $$=gc3(NIL);|}
	  | impDecl			{|imps = singleton($1); $$=gc1(NIL);|}
	  ;
chase	  :     			{if (chase(imps)) {
					     clearStack();
					     onto(imps);
                                             done();
					     closeAnyInput();
					     return 0;
					 }
					 $$ = gc0(NIL);
					}
	  ;
impDecl   : IMPORT modid impspec        { addQualImport($2,$2);
					 addUnqualImport($2,$3);
                                         $$ = gc3($2); }
	  | IMPORT modid ASMOD modid impspec
                                        { addQualImport($2,$4);
					 addUnqualImport($2,$5);
                                         $$ = gc5($2); }
	  | IMPORT QUALIFIED modid ASMOD modid impspec
					{ addQualImport($3,$5);
                                         $$ = gc6($3); }
	  | IMPORT QUALIFIED modid impspec
					{ addQualImport($3,$3);
                                         $$ = gc4($3); }
	  | IMPORT error                {|syntaxError("import declaration");|}
	  ;
impspec   :                             {|$$ = gc0(DOTDOT);|}
	  | HIDING '(' imports ')'      {|$$ = gc4(ap(HIDDEN,$3));|}
	  | '(' imports ')'             {|$$ = gc3($2);|}
	  ;
imports   : 		                {|$$ = gc0(NIL);|}
	  | ','                         {|$$ = gc1(NIL);|}
	  | imports1                    {|$$ = $1;|}
	  | imports1 ','                {|$$ = gc2($1);|}
	  ;
imports1  : imports1 ',' import         {|$$ = gc3(cons($3,$1));|}
	  | import                      {|$$ = gc1(singleton($1));|}
	  ;
import    : var                         {|$$ = $1;|}
	  | CONID                       {|$$ = $1;|}
	  | CONID '(' UPTO ')'          {|$$ = gc4(pair($1,DOTDOT));|}
	  | CONID '(' names ')'         {|$$ = gc4(pair($1,$3));|}
	  ;

@ @<Additional \CEE/ code @>=
/* Omitted. */

@* Index.
