% This file is part of YWEB package.
% --------------------------------------------------------------
% $Source: /home/web/cweb-iso8859/yweb/examples/RCS/rpcalc.w,v $
% $Date: 1998/04/17 17:35:30 $
% $Revision: 1.2 $
% --------------------------------------------------------------

\def\title{RPCALC (Version 1.0)}
\def\topofcontents{\null\vfill
  \centerline{\titlefont Reverse Polish Notation Calculator}
  \vskip 15pt
  \centerline{(Version 1.0)}
  \vfill}
\def\botofcontents{\vfill
\noindent
Copyright \copyright\ 1998 W{\l}odek Bzyl\par}
\pageno=\contentspagenumber \advance\pageno by 1
\let\maybe=\iftrue

@* Introduction.
This is the very first literate program written in \.{YWEB} --- a
literate system for \YACC/ programming.  The literate source for this
program is named `\.{rpcalc.w}'. 

The command \.{ytangle rpcalc.w} creates an output file \.{rpcalc.y},
ready to be processed by an \YACC/ compiler. The command \.{yweave
rpcalc.w} creates an output file \.{rpcalc.tex}, which can then be
processed by \TeX, yielding a ``pretty printed'' version \.{rpcalc.dvi}
which correctly handles typographic details of \YACC/ code.

@f line YACC 
   /* \&{line} is \CPLUSPLUS/ reserved word; 
      we want to typeset it as \YI{line} */

@ The overall layout of a \YACC/ source.  
The `\.{\%\%}', `\.{\%\LB}', \.{\%\RB} are punctuation that appears in
every \YACC/ grammar file to separate the sections.

@c
%{
@<\CEE/ declarations@>
%}
@<\YACC/ declarations@>
%%
@<Grammar rules and actions@>
%%
@<Additional \CEE/ code@>

@ The \CEE/ declarations may define types and variables used in the
actions.  The type of semantic values in calculator grammar is
specified by the macro |YYSTYPE|. Because we specified |double| as the
type of semantic values, each token and each expression has an
associate value, which is a floating number.

@<\CEE/ declarations@>=
#include <stdio.h>
#include <math.h>
#define YYSTYPE double

@ The \YACC/ declaration declare the names of the terminal and
nonterminal symbols, and may also describe operator precedence and the
data types of semantic values of various symbols.

@<\YACC/ declarations@>=
%token NUM

@ The grammar rules define how to construct each nonterminal symbol
from its parts.

@<Grammar rules and actions@>=
input: /* empty */
  | input line 
  ;
line: '\n'
  | exp '\n' {@+printf("\t%.10g\n", $1);@+}
  ;
exp: NUM {|$$=$1;|}
  | exp exp '+' {|$$=$1+$2;|}
  | exp exp '-' {|$$=$1-$2;|}
  | exp exp '*' {|$$=$1*$2;|}
  | exp exp '/' {|$$=$1/$2;|}
  | exp exp '^' {|$$=pow($1,$2);|} /* exponentiation */
  | exp 'n' {|$$=-$1;|} /* unary minus */
  ;

@ The additional \CEE/ code can contain any \CEE/ code we want to use.
Often the definitions of the lexical analyzer \.{yylex} goes here,
plus subroutines called by the actions in the grammar rules. 
In a simple program, as this one, all the rest of the program goes
here.

@ Only a simple lexical analyzer is needed for the RPN calculator.
The lexical analyzer passes a double floating point number via
|yylval| and the token |NUM|, or the ASCII character read if not a
number, skips all blanks and tabs, returns 0 for |EOF|.

@<Additional \CEE/ code@>=
#include <ctype.h>
yylex()
{
  int c;
 
  @<Skip white space@>@;
  @<Process numbers@>@;
  @<Return when |EOF| hit or return single character@>@;
}

@ @<Skip white space@>=
while ((c=getchar())==' ' || c=='\t') ;

@ @<Process numbers@>=
if (c=='.' || isdigit(c)) {
  ungetc(c,stdin);
  scanf("%lf",&yylval); /* global variable defined in \YACC/ */
  return NUM;
}

@ Any character different from `\.{+}', `\.{-}', `\.{*}', `\.{/}',
`\.{\^}' and \.{n} will results in syntax error.

@<Return when |EOF| hit or return single character@>=
if (c==EOF) return(0);
return(c);

@ When |yyparse()| detects a syntax error, it calls the error
reporting function |yyerror()| to print an error message.

@<Additional \CEE/ code@>=
yyerror(s)
  char *s;
{
  printf("%s\n",s);
}

@ The controlling function is kept to bare minimum.  The only
requirement is that it calls |yyparse()| to start the process of
parsing.

@<Additional \CEE/ code@>=
main ()
{
  yyparse();
}

@ Index.  Here is a list of all uses of all identifiers, underlined at
the point of definition. Single-letter identifiers are indexed only
when defined.
